/****************************************************************************
 Module
   Drive_SpeedControl.c

 Description
   Motor speed control module for 218b project drive motors, 
	 simplified to implement a cruise control algorhythm for ENGR 105

	 Controller consists of a cascaded distance and speed controller.
		Distance conroller is PD.  It accepts an input distance and outputs
		a desired RPM.  This desired RPM is the input to the PI RPM controller,
		which outputs a PWM duty cycle to command a percentage of full battery
		voltage.

 Modifications
   1. Set SetClampRPM() function implemented to make the distance controller
			rail immediately, removing the distance controller's ability to
			dynamically change the desiredRPM.  This desiredRPM is then a step
			reference input to the cruise control RPM controller.
	 
	 2. UPDATE_TIME increased to avoid control loops with 0 speed.
			If fewer than 2 encoder ticks are measured between loops, 0 speed
			is recorded.
			
	 3. Kp and Ki values adjusted for experiment from Kp = 2 and Ki = .1,
			which were used in the 218 robot.
			
	 4. CONTROLVARS array defined to fill with data meaningfuol to 105 and
			Getter function written to access this array from the test module.
			
	 5. Commented to presentation quality

 History
 When           Who     What/Why
 -------------- ---     --------
 02/16/16 15:05 lxw		  first pass
 02/17/16 			John A  write distance / RPM controller
 03/06/16				John A  modify for ENGR 105
****************************************************************************/
/*----------------------------- Include Files -----------------------------*/
/* include header files for the framework and this service
*/

// Basic framework modules
#include "ES_Configure.h"
#include "ES_Framework.h"
#include "ES_DeferRecall.h"
#include "ES_ShortTimer.h"

// Driving oriented modules
#include "Drive_SpeedControl.h"
#include "Drive_Motor.h"
#include "Enc_Sense.h"
#include "Drive_SpeedUpdate_Timer.h"
#include "Campaign_MasterMachine.h"

// Specialize framework tools and definitions
#include "inc/hw_timer.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_gpio.h"
#include "inc/hw_sysctl.h"
#include "driverlib/sysctl.h"
#include "driverlib/pin_map.h"	// Define PART_TM4C123GH6PM in project
#include "driverlib/gpio.h"
#include "BITDEFS.h"

/*----------------------------- Module Defines ----------------------------*/
// Readability defines:
#define ALL_BITS		(0xff<<2)

// Codes to define which motor and which encoder data and instruction relate to.
#define LEFT_A 1
#define RIGHT_A 2
#define LEFT_B 3
#define RIGHT_B 4
#define BOTH	0

// TIVA microcontroller physical constants
#define TICKS_PER_SECOND	 40000000
#define TICKS_PER_MS 40000

// Igarashi motor and Pololu encoder constants
#define GEAR_RATIO 50
#define PULSES_PER_REV 5

// Some threshold constants to keep us from micromanaging our position error.
#define MIN_DISTANCE_ERROR	5
#define MIN_TICKS	0

// Controller gains for the distance controller
//		Meaningless for 105 experiment - desired speed clamped to 60 RPM
#define DIST_P_GAIN	5
#define DIST_I_GAIN	0
#define DIST_D_GAIN 150

// Controller gains for speed controller
#define RPM_P_GAIN	.684 // Kp = 1 V/(rad/sec)*(1min/60sec)*(2pi/REV)*(100%/15.3V) = .684 %/RPM
#define RPM_I_GAIN	68.4 // Ki = A*Kp.  Data taken with A = 100 and A = .1 

// Should one wheel turn oppositely?
#define STRAIGHT	0 // both wheels same direction
#define TURN			1 // right wheel goes opposite to left

// How often should our controller update the duty cycle?
#define UPDATE_TIME 	5			//2msec /John changed
#define MaxRunCount 7000    // How many iterations of data do we want to capture?

/*---------------------------- Module Functions ---------------------------*
  prototypes for private functions for this service.  They should be functions
  relevant to the behavior of this service */

//static void SetRPM(uint8_t wheel, float newSpeed);
static float Clamp(float, float, float);

/*---------------------------- Module Variables ---------------------------*/

//Data private to the module
// Distance in encoder ticks.  Speed in RPM.
static float DesiredDistance; 			// input to Dist Controller
static float DesiredSpeed;					// output of Dist controller
static float DistError;							// Dist controller error
static float LastDistError = 0;			// Last time's error
static float DistErrorIntegral = 0;	// Accumulated distance error // NOT USED
static float DistErrorDeriv;				// Differnce between this and last time's e
static int LastTickCount_1;					// How far has left gone?
static int LastTickCount_2;					// How far has right gone?

static float DesiredSpeed_1;				// Input to Left speed controller
static float DesiredSpeed_2;				// Input to Right speed controller
static float LastRecordedSpeed_1;		// Left motor speed measured from encoder
static float LastRecordedSpeed_2;		// Right motor speed measured from encoder
static float RequestedDuty_1;				// Output of left motor speed conrol
static float RequestedDuty_2;				// Output of right motor speed conrol

//speed control loop error variables
static float RPMError_1;
static float RPMError_2;
static float IntegralTerm_1 = 0.0;
static float IntegralTerm_2 = 0.0;

// put data in here to export out to excel later
static float CONTROLVARS [5][MaxRunCount/UPDATE_TIME];

// index to use for throwing data into CONTROLVARS
static uint32_t runCount=0;

// UNUSED. NOW TAKE NEGATIVE DISTANCE TO SET DIRECTION
//static bool Reverse_1;
//static bool Reverse_2;

// Some status codes
static uint8_t DriveMode;
static bool Driving;

// Top speed distant controller will put out
static float ClampRPM;

//static uint32_t checkSpeed_2;

/*------------------------------ Module Code ------------------------------*/

uint32_t GetRunCount(void){
	return runCount;
}
float GetDuty1(void){
	return RequestedDuty_1;
}
float GetDuty2(void){
	return RequestedDuty_2;
}
float checkSpeed(void){
	return LastRecordedSpeed_2;
}
float GetDesiredSpeed2(void) {
	return DesiredSpeed_2;
}
float GetIntegral2(void) {
	return IntegralTerm_2;
}
// Combines the above to get data at the end without slowing down controller
float GetControlVar(int p, int t){
	return CONTROLVARS[p][t];
}
/****************************************************************************
 Function
     Drive_SpeedControl_Init
 Parameters
     void
 Returns
     void
 Description
     Initializes speed control for drive motors
 Author
     LXW, 02/16/16, 15:20
****************************************************************************/
void Drive_SpeedControl_Init(void){
	 //initialize drive motor control
	Drive_Motor_Init();
	
	 //initialize drive motor encoders
	Enc_Sense_Init();
	
	 //initialize the periodic speed update timer
	Drive_SpeedUpdate_Timer_Init(UPDATE_TIME);
	
	printf("\r\nSPEED CONTROL INITIALIZED\n");
}
// Tell motor controllers to stop at the current position
void Drive_Stop(void){
	Driving = false;
	//reset control variables
	DesiredDistance = 0;
	DesiredSpeed = 0;
	DesiredSpeed_1 = 0;
	DesiredSpeed_2 = 0;
	Enc_ResetTickCount(BOTH);
	DistErrorIntegral = 0;
	DistErrorDeriv = 0;
	LastTickCount_1 = 0;
	LastTickCount_2 = 0;
	//set new distance setpoint
	DesiredDistance = 0;
}
// Tell motor controllers to achieve the newLimit distance
void Drive_SetDistance(float newLimit){
	//reset control variables
	DesiredDistance = 0;
	DesiredSpeed = 0;
	DesiredSpeed_1 = 0;
	DesiredSpeed_2 = 0;
	Enc_ResetTickCount(BOTH);
	DistErrorIntegral = 0;
	DistErrorDeriv = 0;
	LastTickCount_1 = 0;
	LastTickCount_2 = 0;
	//set new distance setpoint
	DesiredDistance = newLimit;
	Driving = true;
	runCount = 0; //for purpose of starting to collect data here!
}
/****************************************************************************
 Function
   Drive_SetMode
 Parameters
	uint32_t : new driving mode (Turning or Driving straight
 Returns
   void
 Description
	set new drive mode
 Author
   lxw, 02/17/16, 13:50
****************************************************************************/
void Drive_SetMode(uint8_t newMode){
	DriveMode = newMode;
}
/****************************************************************************
 Function
   Drive_SetClampRPM
 Parameters
	float : new clamping rpm
 Returns
   void
 Description
	set new clamp RPM
	For 105: this clamp is set to 60 RPM to induce a step input reference speed
 Author
   lxw, 02/17/16, 13:50
****************************************************************************/
void Drive_SetClampRPM(float newRPM){
	ClampRPM = newRPM;
}
/****************************************************************************
 Function
   Drive_GetRPM
 Parameters
	uint8_t : selected wheel
 Returns
	uint16_t : last measured RPM of selected wheel
 Description
	return last recorded RPM for a drive wheel   
 Author
   lxw, 02/16/16, 16:15
****************************************************************************/
float Drive_GetRPM(uint8_t wheel){
	if(wheel == LEFT_A){
		return LastRecordedSpeed_1;
	}
	else if(wheel == RIGHT_A){
		return LastRecordedSpeed_2;
	}
	else{
		return 0;
	}
}
/****************************************************************************
 Function
   Drive_SpeedControlISR
 Parameters
	 void
 Returns
   void
 Description

	 This is primary workhorse of this module.  It fires on interrupt every
	 UPDATE_TIME msec.  Each loop, it runs 2 cascaded controllers.  The first
	 is a PD controller on distance to travel.  Input is DesiredDistance; output
	 is DesiredSpeed.  Next, DesiredSpped is broken up into DesiredSpeed 1 and 2,
	 one for each left and right motor.  These serve as inputs to the PI speed
	 controller, which output RequestedDuty cycles 1 and 2.  These duty cycles
	 run a PWM voltage to the motor drivers.  Negative values send a bit to the
	 reverse register on the motor driver.

 Authors
   lxw, 01/28/16, 19:00
	 John A, 02/17 and 03/06
****************************************************************************/
void Drive_SpeedControlISR(void){
	
	 //start by clearing the source of the interrupt
	HWREG(WTIMER1_BASE+TIMER_O_ICR) = TIMER_ICR_TATOCINT;
	
	//****************************************************************NEW INFO GATHERING
	//for wheel 1 (LEFT)
	 //if not enough new ticks have not been found
	if(abs(Enc_GetTickCount(1) - LastTickCount_1) <= MIN_TICKS){
		 //set last recorded encoder RPM to 0
		LastRecordedSpeed_1 = 0;
	}
	 //else
	else{
		 //calculate RPM from current period
		LastRecordedSpeed_1 = ((TICKS_PER_SECOND/(Enc_GetPeriod(LEFT_A))*60)/(PULSES_PER_REV*GEAR_RATIO));
		 //capture new tick count
		LastTickCount_1 = Enc_GetTickCount(1);
	}
	//for wheel 2 (RIGHT)
	 //if not enough new ticks have not been found
	if(abs(Enc_GetTickCount(2) - LastTickCount_2) <= MIN_TICKS){
		 //set last recorded encoder RPM to 0
		LastRecordedSpeed_2 = 0;
	}
	 //else
	else{
		 //calculate RPM from current period
		LastRecordedSpeed_2 = ((TICKS_PER_SECOND/(Enc_GetPeriod(RIGHT_A))*60)/(PULSES_PER_REV*GEAR_RATIO));
		 //capture new tick count
		LastTickCount_2 = Enc_GetTickCount(2);
	}
	
	//****************************************************************POSITION CONTROL
	
	//Calculate Distance Error
	 //if driving straight
	if(DriveMode == STRAIGHT){
		DistError = DesiredDistance - (LastTickCount_1+LastTickCount_2)/2; //taking average
	}
	 //else if turning
	else{
		DistError = DesiredDistance - (-LastTickCount_1+LastTickCount_2)/2; //taking average with inverse of motor 1
	}
	
	// Calc Dist Error Integral
	DistErrorIntegral += DistError;
	DistErrorIntegral = Clamp(DistErrorIntegral, -ClampRPM, ClampRPM); /* anti-windup */

	//Calculate Distance Error Derivative
	DistErrorDeriv = DistError - LastDistError;
	
	//Set Last Distance Error for next time
	LastDistError = DistError;
	
	//Apply PD
	DesiredSpeed = DIST_P_GAIN*DistError + DIST_I_GAIN*DistErrorIntegral + DIST_D_GAIN*DistErrorDeriv;
	DesiredSpeed = Clamp(DesiredSpeed, -ClampRPM, ClampRPM); // Clamp set low by test command.
																													 // All outputs should clamp for 105 tests!
	//Conditional based on drive mode
	//if currently driving straight
	if(DriveMode == STRAIGHT){
		DesiredSpeed_1 = DesiredSpeed;
		DesiredSpeed_2 = DesiredSpeed;
	}
	 //else if turning
	else{
		DesiredSpeed_1 = -DesiredSpeed;
		DesiredSpeed_2 = DesiredSpeed;
	}
	
	//****************************************************************TICK COUNT MONITOR
	// Have we achieved our desired distance?  If so, stop!
	if( (abs(DistError) < MIN_DISTANCE_ERROR) && (abs(LastDistError) < MIN_DISTANCE_ERROR) && Driving == true){
		
		Driving = false;
		//post event up to TestService to signal next step (for checkoff and testing only)
		ES_Event doneEvent;
		doneEvent.EventType = ES_MOVE_COMPLETE;
		PostCampaignService(doneEvent);
	}
 
	//****************************************************************SPEED CONTROL FOR LEFT MOTOR (1)
	
	//Not modified for 105.  Only measuring right motor!  See better comments for right motor
	RPMError_1 = DesiredSpeed_1 - LastRecordedSpeed_1; // In RPM
	IntegralTerm_1 += RPMError_1;
	IntegralTerm_1 = Clamp(IntegralTerm_1, -100, 100); /* anti-windup */
	RequestedDuty_1 = RPM_P_GAIN*RPMError_1 + RPM_I_GAIN*IntegralTerm_1;
	RequestedDuty_1 = Clamp(RequestedDuty_1, -100, 100);
	Drive_SetDuty_Left(RequestedDuty_1); // output calculated control
	
	//****************************************************************SPEED CONTROL FOR RIGHT MOTOR (2)
	
	RPMError_2 = DesiredSpeed_2 - LastRecordedSpeed_2; // Calculate RPM error for this loop
	
	IntegralTerm_2 += RPMError_2; // Increment the integral of all recorded RPM errors for right motor.
	
	IntegralTerm_2 = Clamp(IntegralTerm_2, -100/RPM_I_GAIN, 100/RPM_I_GAIN); /* anti-windup */
																	// Scaled to Gains for 105 experiment.
																	// This should more succesfully keep efforts below 100% duty cycle
	
	RequestedDuty_2 = RPM_P_GAIN*RPMError_2 + RPM_I_GAIN*IntegralTerm_2;  // Add both efforts together
	RequestedDuty_2 = Clamp(RequestedDuty_2, -100, 100); // Just in case.
	
	Drive_SetDuty_Right(RequestedDuty_2); // output calculated control to duty cycle module
	
	// increment index
	runCount++;
	// record data
	if (runCount <= MaxRunCount/UPDATE_TIME){
		CONTROLVARS[0][runCount-1] = (runCount-1)*UPDATE_TIME;	// in msec
		CONTROLVARS[1][runCount-1] = DesiredSpeed_2;						// in RPM
		CONTROLVARS[2][runCount-1] = LastRecordedSpeed_2;
		CONTROLVARS[3][runCount-1] = IntegralTerm_2;
		CONTROLVARS[4][runCount-1] = RequestedDuty_2;						// in %
	}
}
/****************************************************************************
 Function
  clamp
 Parameters
	3x float
 Returns
	input value, clamped between upper and lower input bounds  
 Author
	LxW, 01/28/16, 21:00
****************************************************************************/
static float Clamp(float x, float lower, float upper){
	float y;
	if(x >= upper){
		y = upper;
	}
	else if(x <= lower){
		y = lower;
	}
	else{
		y = x;
	}
	return y;
}
